/*
 *
 *
 *    RttyModem.ino
 *
 *    A simple terminal/modem for RTTY.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2016-2018 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

//
//   NOTES:
//
//     1. This firmware WILL write to the EEPROM area of the Teensy board.  It
//        will also attempt to wear-level the writes to prevent premature wear
//        of the EEPROM.
//     2. If upgrading from a previous version, you should execute the command
//        #EC=1; from the serial port to clear the previous EEPROM contents.
//     3. Several configuration options follow, and are marked either REQUIRED,
//        OPTIONAL, or ADVANCED.  The values provided are reasonable defaults,
//        and you should avoid changing any configuration options unless you
//        are sure of the changes that will result.
//

//
//  REQUIRED: This setting controls the center frequency (in Hz)
//
//  This can be any (reasonable) value, but there are some fairly standardized
//  AFSK tone frequencies.  When combined with an RTTY_DATA_FREQ_SHIFT of 170:
//     A value of 2210 will give a 2125/2295 tone pair (the "standard" high pair).
//     A value of 1500 will give a 1415/1585 tone pair.
//     A value of 1360 will give a 1275/1445 tone pair (the "standard" low pair).
//     A value of 1000 will give a 915/1085 tone pair.
//
#define RTTY_DATA_FREQ_CENTER (1360)

//
//  REQUIRED: The frequency shift.
//
//  Likewise, this calculates the shift between tones; one half of
//  this value will be added or subtracted from the CENTER frequency
//  defined above, to generate AFSK output; setting is in Hz.
//
//  For amateur radio use, leave this set to 170.
//
#define RTTY_DATA_FREQ_SHIFT (170)

//
//  REQUIRED: The bit rate.
//
//  For amateur radio use, leave this set to 45.45.
//
#define RTTY_DATA_BIT_RATE (45.45)

//
//  REQUIRED: DATA pin assignment.
//
//  Set the DATA output pin; this pin sends logic levels to a transciever's
//  FSK input.  By default, the polarity is configured to drive an NPN transistor,
//  or an n-channel FET, that will ground the FSK line from the radio.  This
//  can be modified at runtime with the #REV command.
//
#define RTTY_KEY_PIN (2)

//
//  REQUIRED: PTT pin assignment.
//
//  Set the PTT output pin; this pin will be driven LOW to activate the
//  radio PTT through an NPN transistor or n-channel FET.
//
#define RTTY_PTT_PIN (3)

//
//  OPTIONAL: Heartbeat LED
//
//  The optional heartbeat LED blinks at 1Hz to show normal loop activity
//  Set to zero (0) to disable.
//
#define HEARTBEAT_LED (0)

//
//  OPTIONAL: LED trigraph pin assignments
//
//  Set these to enable an LED tuning indicator attached to discrete LEDs.
//  Set all to zero to disable
//
#define RTTY_TRIGRAPH_LEFT (5)
#define RTTY_TRIGRAPH_CENTER (6)
#define RTTY_TRIGRAPH_RIGHT (7)

//
//  REQUIRED: Host port assignment
//
//  Assign a specific hardware port to be the interface to the host; by
//  default, the USB connection on most boards is 'Serial'
//
#define HostSerial Serial

//
//  REQUIRED: Host port speed
//
//  Set the 'baud rate' of the serial port.
//
#define SerialRate (9600)

//
//  REQUIRED: PTT to KEY-up delay
//
//  Set the delay (in msec) between releasing PTT and releasing the KEY
//
#define PTT_TO_UNKEY_DELAY (5)

//
//  ADVANCED: Set the sample rate divisor
//
//  This is the downsampling factor to run after the IIR preselector.  Values
//  between 1 and 8 are appropriate for AF frequencies between 400 and 2000Hz.
//
//  Different divisors will produce different effective sampling rates for
//  the detector algorithm.  The preselector will run at the codec's 44100Hz,
//  regardless of this setting.
//
//      Divisor     Fs
//         1        44100
//         2        22050
//         3        14700
//         4        11025
//         5        8820
//         6        7350
//         7        6300 (recommended default)
//         8        5512.5
//
#define RTTY_RATE_DIVISOR (7)

//
//  REQUIRED: Set the input volume level.
//
//  This controls the input volume in the Teensy codec, to set the level
//  corresponding to 100% volume.  Valid values are 0 to 15, and correspond
//  to the following levels on the Teensy audio shield:
//
//   0: 3.12 Volts p-p
//   1: 2.63 Volts p-p
//   2: 2.22 Volts p-p
//   3: 1.87 Volts p-p
//   4: 1.58 Volts p-p
//   5: 1.33 Volts p-p
//   6: 1.11 Volts p-p
//   7: 0.94 Volts p-p
//   8: 0.79 Volts p-p
//   9: 0.67 Volts p-p
//   10: 0.56 Volts p-p
//   11: 0.48 Volts p-p
//   12: 0.40 Volts p-p (recommended default)
//   13: 0.34 Volts p-p
//   14: 0.29 Volts p-p
//   15: 0.24 Volts p-p
//
//  This setting only represents the default value.  The value can be changed
//  by using the LINEIN command.
//
#define RTTY_INPUT_LEVEL (12)

//
//  ADVANCED: set the sharpness of the individual tone filters
//
#define RTTY_TONE_FILTER_R (0.977)

//
//  ADVANCED: set the length of the individual tone integrators
//
//  This value is measured in cycles, and the actual integrator lengths are
//  computed individually based on this value.
//
#define RTTY_TONE_INTEGRATOR_LENGTH (20)

//
//  ADVANCED: set the hysteresis limit for the discriminator
//
//  This value is normalized, and represents the fraction of the absolute
//  maximum value of the angle between the tone magnitudes.  This value
//  determines the hysteresis thresholds used to prevent the discriminator
//  from "chattering" when the deviation value is near zero.
//
#define RTTY_DISCRIMINATOR_HYSTERESIS_TRIGGER (0.020)

//
//  ADVANCED: the number of audio buffer frames to allocate
//
#define RTTY_AUDIO_BUFFER_FRAMES (20)

//
//  ADVANCED: Set the bandwidth of the input preselector/decimator
//            filter, measured in Hz.
//
//  Values less than 500 are not recommended, because the IIR filter
//  can start to ring on background noise.  This need not be less
//  than 500Hz, because the radio's narrow filter should provide any
//  additional selectivity desired.  This value is enough to make the
//  algorithms function properly, even on an open 3kHz SSB channel.
//
//  Values larger than 500 may actually help in the presence of certain
//  kinds of noise.
//
//  The preselector will also serve as an antialiasing filter, allowing
//  the detector to run at a lower sampling frequency, which will require
//  less CPU processing power.
//
#define RTTY_BW (500.0)

//
//  ADVANCED: Set the order of the input filter.
//
//  Zero will disable the filter entirely, valid values are zero through four.
//  Higher values will require more CPU time, but will make the transition
//  bands sharper.
//
//  A value of one is recommended for initial testing.
//
#define RTTY_ORDER (1)

//
//  DEPRECATED: invert tone polarity.
//
//  Set this to make the 'normal' polarity for both TX and RX inverted.
//  This also inverts all values for the #REV command's settings.
//  Leaving this commented out matches the standard tone pairs on an
//  LSB transciever using an n-channel or NPN transistor for keying
//  the transmitter.
//
//  It is recommended that you leave this DISABLED, and adjust the
//  polarity as needed using the #REV command.
//
//#define RTTY_INVERTED

// ================== END OF CONFIGURATION SETTINGS ==================

//
//  ...everything below this point is the sketch code; do not change
//     anything below this unless you know what you are doing...
//

#if RTTY_RATE_DIVISOR > 1 && RTTY_ORDER == 0
#error Use of downsampling requires the preselector order of >= 1
#endif

// Fs is fixed on the audio adaptor
#define RTTY_SAMPLE_RATE (44100 / RTTY_RATE_DIVISOR)

// this is the packet size returned by the codec
#define SG_PACKET_SIZE (128)

// include Teensy headers for audio
#include <Audio.h>
#include <Wire.h>

//
//  Teensy Audio Library Configuration
//
//  NOTE: this configuration reads the LEFT channel only
//        from the audio shield, to obtain mono audio.
//

// GUItool: begin automatically generated code
AudioControlSGTL5000     sgtl5000;       //xy=805,54
AudioInputI2S            i2sIn;          //xy=110,142
AudioFilterBiquad        iirIn;          //xy=560,145
AudioRecordQueue         inQueue;        //xy=807,146
AudioAnalyzePeak         peakIn;          //xy=540,254
AudioConnection          patchCord1(i2sIn, 0, iirIn, 0);
AudioConnection          patchCord2(iirIn, inQueue);
AudioConnection          patchCord3(iirIn, peakIn);
// the following will be used for AFSK output in a future release
//AudioPlayQueue           outQueue;       //xy=109,235
//AudioFilterBiquad        iirOut;         //xy=341,235
//AudioOutputI2S           i2sOut;         //xy=804,230
//AudioConnection          patchCord4(outQueue, iirOut);
//AudioConnection          patchCord5(iirOut, 0, i2sOut, 0);
//AudioConnection          patchCord6(iirOut, 0, i2sOut, 1);
// GUItool: end automatically generated code

// local headers
#include "CircularBuffer.h"
#include "Elapsed.h"
#define HEARTBEAT_ACTIVE (200000)
#define HEARTBEAT_IN_MICROS
#include "Heartbeat.h"
#include "SplitCommand.h"
#include "baudot.h"
#include "tpfdisc.h"
#include "sign.h"
#include "nlimits.h"
#include "eeblock.h"
#include "enums.h"
#include "settings.h"

// the version string
#define VERSION_STRING ("RttyModem 2.0 (beta-20180228d)")

// CR+LF
#define LOCAL_CRLF ("\r\n")

// import some namespaces
using namespace KK5JY::Collections;

//
//  Constants
//

// time limits
#define MaxPttLeadIn (1000L)		// msec
#define MaxPttLeadOut (1000L)		// msec

// max string lengths
#define InputStringLength (128)
#define CommandStringLength (128)
#define TxBufferLength (128)
#define IoBufferLength (128)

// control keys
#define ESC '\x1B'
#define CtlC '\x03'

// define *default* RTTY polarities
#ifdef RTTY_INVERTED
#define RTTY_HIGH HIGH
#define RTTY_LOW LOW
#else
#define RTTY_HIGH LOW
#define RTTY_LOW HIGH
#endif

////////////////////////////////////////////////////////////////
//
//  Globals
//
////////////////////////////////////////////////////////////////

// TX buffer - each element is an encoded byte with start/stop bits added
CircularBuffer<uint16_t> TxBuffer(TxBufferLength);
// TX shift register vars
uint16_t TxMask = 1;

// RX shift register vars
uint16_t RxByte = 0;
uint16_t RxMask = 1;

// a string to hold incoming data
String inputString;

// and another for command data
String cmdText = "";

// command parsing
String command = "";
String argument = "";

// whether the string is complete
bool inputReady = false;

// I/O buffer
char ioBuffer[IoBufferLength];

// echo USB RX data back to host
uint8_t echo = false;

// currently parsing a command
bool cmd = false;

// TX mode (otherwise RX)
bool tx = false;

// tracks PTT state without reading PTT pin
#if RTTY_PTT_PIN > 0
bool pttState = false;
#endif

// tracks the shifted state of the transmitter or receiver
bool txShift = false;
bool txFresh = true;
bool rxShift = false;
bool rxFresh = true;
bool uos = true; // UOS setting for receiver

// current value of RX pin (only assigned by ISR)
volatile uint8_t rttyRxState = false;

// last value of RX pin considered (used for state change detection)
uint8_t rttyRxLastState = LOW;

// RX element timing - stores the last state change time
unsigned long rttyRxEdge = 0;

// RX element timing - stores the last clock edge
unsigned long rttyRxClock = 0;

// TX element timing
unsigned long rttyTxEdge = 0;

// global timer
volatile unsigned long now = 0;

// flag indicating that the settings should be written
static bool settingsDirty = false;
static uint32_t lastSettingsWrite = 0;
#define SettingsInterval (10 * 1000) // msec

// the settings block
SettingsBlock Settings;

// the EEPROM key
static const char *EEPromKey = "RTTYMOD_104";

// the EEPROM access
static EEBlock eeprom(EEPromKey);

// holds one bit in the MSB that can be decoded for one byte
uint16_t maxMask = 0;

// holds the stop-bit mask
uint16_t stopBits = 0;

// the bit period (in usec)
uint16_t bitPeriod = 0;

// bit envelope limits
uint32_t maxPeriod = 0;
uint32_t halfBit = 0;

// the demodulator type
typedef int16_t sample_t;
typedef int32_t accum_t;
typedef float filter_t;
typedef KK5JY::DSP::TwinPassbandDiscriminator<sample_t, accum_t, filter_t> Demodulator;
Demodulator *demod = 0;

// the default squelch setting
static uint32_t squelchOpened = 0;
static bool squelch = true;
#define SQUELCH_TIMEOUT_MAX (5000)

// highest line-in volume
#define LINE_INPUT_MAX (15)

// allocate the trigraph
#if (RTTY_TRIGRAPH_LEFT != 0) && (RTTY_TRIGRAPH_CENTER != 0) && (RTTY_TRIGRAPH_RIGHT != 0)
#define RTTY_TRIGRAPH
#include "TriGraph.h"
KK5JY::TriGraph trigraph(RTTY_TRIGRAPH_LEFT, RTTY_TRIGRAPH_CENTER, RTTY_TRIGRAPH_RIGHT);
static uint8_t triCtr = 0;
#define TRI_TARGET (1) // process at least this many frames between LED updates
#define TRI_LED_EDGE (80)
#endif

// audio clock
static uint32_t audioClock = 0;
static uint8_t audioClockCtr = 0;
#define AUDIO_CLOCK_TARGET (8)
#if RTTY_RATE_DIVISOR != 1
static uint8_t downsample_ctr = 0;
#endif


////////////////////////////////////////////////////////////////
//
//  Support Methods
//
////////////////////////////////////////////////////////////////


// returns true if TX should be inverted
static inline bool ReverseTX() {
	return (Settings.reverse & RevTX) == RevTX;
}

// returns true if RX should be inverted
static inline bool ReverseRX() {
	return (Settings.reverse & RevRX) == RevRX;
}

// LOW state for RX
static inline bool LowRX() {
	return ReverseRX() ? RTTY_HIGH : RTTY_LOW;
}

// HIGH state for RX
static inline bool HighRX() {
	return ReverseRX() ? RTTY_LOW : RTTY_HIGH;
}

// LOW state for TX
static inline bool LowTX() {
	return ReverseTX() ? RTTY_HIGH : RTTY_LOW;
}

// HIGH state for TX
static inline bool HighTX() {
	return ReverseTX() ? RTTY_LOW : RTTY_HIGH;
}

//
//  ReconfigureTiming()
//
static void ReconfigureTiming() {
	// compute the mask value that causes a new byte to be used
	maxMask = GetMaxMask();
	// compute the stop-bit mask
	stopBits = (maxMask >> 1) | maxMask;
	// compute the bit period (usec)
	bitPeriod = round(1000000.0 / RTTY_DATA_BIT_RATE);
	// calculate the max bit envelope
	maxPeriod = 16L * bitPeriod;
	// calculate the minimum bit time threshold
	halfBit = (Settings.halfBitFactor * bitPeriod);

	// clear the shift register
	RxMask = 1;
	RxByte = 0;
}

//
//  ServiceSerialPortInputQueue() - empty the Serial object connected to the host port
//
static inline char ServiceSerialPortInputQueue() {
	// service the serial port (one byte at a time)
	if (HostSerial.available()) {
		// get the new byte
		char inChar = (char)HostSerial.read();
		// add it to the input string
		if (inputString.length() < InputStringLength) {
			inputString += inChar;
			inputReady = true;
		}
		return inChar;
	}
	return -1;
}


//
//  Decode one byte from the wire
//
static inline bool Decode(uint16_t raw, char &ch) {
	// queue up the character for transmission
	switch (Settings.encoding) {
		case Baudot:
			if (raw == RTTY_NUL) {
				ch = 0;
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_CR) {
				ch = '\r';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_LF) {
				ch = '\n';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_SP) {
				ch = ' ';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_LS) {
#ifdef VERBOSE_DEBUG
				HostSerial.write("[LS]");
#endif
				ch = 0; // no printable result
				rxShift = false; // but de-shift the decoder state
				return true; // and return success
			} else if (raw == RTTY_FS) {
#ifdef VERBOSE_DEBUG
				HostSerial.write("[FS]");
#endif

				ch = 0; // no printable result
				rxShift = true; // but up-shift into figures mode
				return true; // and return success
			} else if (rxShift) {
				switch (raw) {
					case RTTY_0: ch = '0'; return true;
					case RTTY_1: ch = '1'; return true;
					case RTTY_2: ch = '2'; return true;
					case RTTY_3: ch = '3'; return true;
					case RTTY_4: ch = '4'; return true;
					case RTTY_5: ch = '5'; return true;
					case RTTY_6: ch = '6'; return true;
					case RTTY_7: ch = '7'; return true;
					case RTTY_8: ch = '8'; return true;
					case RTTY_9: ch = '9'; return true;
					case RTTY_DASH: ch = '-'; return true;
					case RTTY_QUOTE: ch = '\''; return true;
					case RTTY_DOLLAR: ch = '$'; return true;
					case RTTY_BANG: ch = '!'; return true;
					case RTTY_AMP: ch = '&'; return true;
					case RTTY_HASH: ch = '#'; return true;
					case RTTY_PARL: ch = '('; return true;
					case RTTY_PARR: ch = ')'; return true;
					case RTTY_PLUS: ch = '+'; return true;
					case RTTY_SLASH: ch = '/'; return true;
					case RTTY_SEMI: ch = ';'; return true;
					case RTTY_QUES: ch = '?'; return true;
					case RTTY_COMMA: ch = ','; return true;
					case RTTY_FSTOP: ch = '.'; return true;
					case RTTY_COLON: ch = ':'; return true;
					default: break;
				}
			} else { // unshifted
				switch (raw) {
					case RTTY_A: ch = 'A'; return true;
					case RTTY_B: ch = 'B'; return true;
					case RTTY_C: ch = 'C'; return true;
					case RTTY_D: ch = 'D'; return true;
					case RTTY_E: ch = 'E'; return true;
					case RTTY_F: ch = 'F'; return true;
					case RTTY_G: ch = 'G'; return true;
					case RTTY_H: ch = 'H'; return true;
					case RTTY_I: ch = 'I'; return true;
					case RTTY_J: ch = 'J'; return true;
					case RTTY_K: ch = 'K'; return true;
					case RTTY_L: ch = 'L'; return true;
					case RTTY_M: ch = 'M'; return true;
					case RTTY_N: ch = 'N'; return true;
					case RTTY_O: ch = 'O'; return true;
					case RTTY_P: ch = 'P'; return true;
					case RTTY_Q: ch = 'Q'; return true;
					case RTTY_R: ch = 'R'; return true;
					case RTTY_S: ch = 'S'; return true;
					case RTTY_T: ch = 'T'; return true;
					case RTTY_U: ch = 'U'; return true;
					case RTTY_V: ch = 'V'; return true;
					case RTTY_W: ch = 'W'; return true;
					case RTTY_X: ch = 'X'; return true;
					case RTTY_Y: ch = 'Y'; return true;
					case RTTY_Z: ch = 'Z'; return true;
					default: break;
				}
			}

#ifdef VERBOSE_DEBUG
			HostSerial.print("#debug,Could not decode symbol: ");
			HostSerial.print(raw);
			HostSerial.println(";");
#endif
			return false;

		case ASCII7:
			raw = ~raw;
			raw >>= 1;
			ch = raw & 0x7F;
			return true;
		case ASCII8:
			raw = ~raw;
			raw >>= 1;
			ch = raw & 0x7F;
			return true;
		default:
			return false;
	}
}


//
//  ProcessShiftRegister() - common code for both reciever's shift register cycle
//
static inline bool ProcessShiftRegister(uint8_t bit) {
	bool result = false;

	// if the input is HIGH, set the LSB
	if (bit == HIGH) {
		RxByte |= RxMask;
	}

	// left-shift to next position
	if (RxMask < maxMask) {
		RxMask <<= 1;
	} else {
		//
		// check for decoded byte
		//
		// at this point, we should have one complete byte,
		//    stop/start bits and all;
		// if so, decode;
		// if not, shift to the first candidate start bit and move on
		//
		if ((RxByte & 0x0001) && ((RxByte & stopBits) == 0x0000)) {
			// valid byte!  Decode and print...
			char ch = 0;
			if ((result = Decode(RxByte, ch))) {
				// if the character can be printed, print it
				switch (ch) {
					case 0: break;
					case '#': {
						if (!squelch)
							HostSerial.print("#HASH;");
					} break;
					default: {
						if (!squelch)
							HostSerial.print(ch);
					} break;
				}

				// reset shift register
				RxByte = 0;
				RxMask = 1;
			}
		}

		// if the shift register wasn't cleared by a successful decode,
		//    shift it over by one bit to allow the contents to be
		//    reused in successive attempts
		if (RxByte)
			RxByte >>= 1;
	}

	return result;
}


//
//  ProcessRxQueue() - RX audio data handler for Teensy audio library;
//                     this is not an interrupt handler, it only processes
//                     the samples in the audio queue...
//
static void ProcessRxQueue() {
	//bool decoded = false;

	while (inQueue.available()) {
		// read one audio frame
		int16_t *data = inQueue.readBuffer();
		int16_t * const target = data + SG_PACKET_SIZE;

		// process each sample within that frame
		while (data != target) {
			int16_t sample = *data++;

			// run the audio clock on the 44.1kHz input stream; this emulates a 1MHz clock, recorded in 'audioClock'
			if (++audioClockCtr == AUDIO_CLOCK_TARGET) {
				audioClockCtr = 0;
			}
			audioClock += ((audioClockCtr == 0) || ((audioClockCtr & 0x01) == 0x01)) ? 23 : 22;

			#if RTTY_RATE_DIVISOR != 1
			// downsample, if enabled
			if (++downsample_ctr >= RTTY_RATE_DIVISOR) {
				downsample_ctr = 0; // process this sample
			} else {
				continue; // skip this sample
			}
			#endif

			// limit/clip the input
			sample =
				sign(sample) == 1 ?
					KK5JY::DSP::norm_limits<sample_t>::maximum :
					KK5JY::DSP::norm_limits<sample_t>::minimum;

			//
			//  RUN: feed the discriminator, and convert to binary
			//
			rttyRxState = demod->run(sample);

			// then convert to bitstream
			uint32_t bitLength = Elapsed(audioClock, rttyRxEdge);
			if (bitLength >= (maxPeriod)) {
				bitLength = maxPeriod;
			}

			// if state changed enough for a complete bit
			//    OR if the line is idle for more than a full byte-time....
			const bool stateChanged = (rttyRxState != rttyRxLastState);
			const bool fullBit = (bitLength >= halfBit);
			const bool tooIdle = (bitLength >= maxPeriod);
			const bool srActive = (RxMask != 1);
			if ((stateChanged && fullBit) || (tooIdle && srActive)) {
				const uint8_t currentState = rttyRxLastState;	// save the previous state
				rttyRxLastState = rttyRxState;					// save the new state (for later edge triggering)

				// save this edge time
				rttyRxEdge = audioClock;

				// at this point we have (length / period) bits of value 'currentState' bits,
				//    rounded to the nearest whole bit
				uint32_t bits = (bitLength / bitPeriod);
				if ((bitLength % bitPeriod) >= halfBit) {
					++bits;
				}

				// for each bit
				for (uint32_t i = 0; i != bits; ++i) {
					// add the bit and process the contents of the shift register
					if (ProcessShiftRegister(currentState == LowRX())) {
						//decoded = true;
					}
				}
			}
		} // ... end process one SAMPLE

		// free the processed frame
		inQueue.freeBuffer();

		// read the current deviation into a signed byte
		int8_t tri = demod->deviation(KK5JY::DSP::norm_limits<int8_t>::maximum);
		if (abs(tri) > Settings.squelch) {
			squelch = false;
			squelchOpened = millis();
		}

		#ifdef RTTY_TRIGRAPH
		// paint the trigraph
		if (triCtr++ == TRI_TARGET) {
			trigraph.update(tri);
			triCtr = 0;
		}
		#endif
	} // ... end process one FRAME
}


//
//  SetRttyState(bool)
//
static inline void SetRttyState(bool bit) {
	// generate logic output
	digitalWrite(RTTY_KEY_PIN, bit ? HighTX() : LowTX());

	// record the transition time
	rttyTxEdge = now = micros();
}

//
//  Get the highest mask value that is valid for current encoding
//
static inline uint16_t GetMaxMask() {
	switch(Settings.encoding) {
		case Baudot:
			// five data bits, no parity, two stop bits
			//       ssddddds
			return 0b10000000;
		case ASCII7:
			// seven data bits, no parity, two stop bits
			//       ssddddddds
			return 0b1000000000;
		case ASCII8:
			// eight data bits, no parity, two stop bits
			//       ssdddddddds
			return 0b10000000000;
	}
	return 0;
}

//
//  Get the value of the next bit to transmit
//
static inline bool GetNextTxBit(bool & bit) {
	// if there is no data, give up
	uint16_t w;
	if (!TxBuffer.First(w))
		return false;

	// compute the result
	bit = (w & TxMask);

	//#if RTTY_INVERTED == true
	//if (TxMask != 1 && TxMask != maxMask && TxMask != (maxMask / 2))
		//bit = !bit;
	//#endif

	// shift the mask to the next bit
	TxMask <<= 1;

	// if we have read all we can get from the current byte...
	if (TxMask > maxMask) {
		// ... go to the next byte
		TxMask = 1;
		TxBuffer.Remove(w);
	}

	// success
	return true;
}


//
//  Enqueue the next byte for transmission
//
static inline int8_t Encode(char ch) {
	// queue up the character for transmission
	switch (Settings.encoding) {
		case Baudot:
			// start with the characters that are shift-independent
			if (ch == 0) {
				return TxBuffer.Add((uint16_t)(RTTY_NUL)) ? 1 : -1;
			} else if (ch == '\r') {
				return TxBuffer.Add((uint16_t)(RTTY_CR)) ? 1 : -1;
			} else if (ch == '\n') {
				return TxBuffer.Add((uint16_t)(RTTY_LF)) ? 1 : -1;
			} else if (ch == ' ') {
				return TxBuffer.Add((uint16_t)(RTTY_SP)) ? 1 : -1;
			} else if (isdigit(ch)) {
				int8_t bc = 0;

				// shift
				if (txFresh || !txShift) {
					TxBuffer.Add((uint16_t)(RTTY_FS));
					txShift = true;
					txFresh = false;
					bc = 1;
				}

				// transform the digit
				switch (ch) {
					case '0': ch = RTTY_0; break;
					case '1': ch = RTTY_1; break;
					case '2': ch = RTTY_2; break;
					case '3': ch = RTTY_3; break;
					case '4': ch = RTTY_4; break;
					case '5': ch = RTTY_5; break;
					case '6': ch = RTTY_6; break;
					case '7': ch = RTTY_7; break;
					case '8': ch = RTTY_8; break;
					case '9': ch = RTTY_9; break;
					default: return 0;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the FS insertion, too

			} else if (isalpha(ch)) {
				int8_t bc = 0;

				// shift
				if (txShift || txFresh) {
					TxBuffer.Add((uint16_t)(RTTY_LS));
					txShift = false;
					txFresh = false;
					bc = 1;
				}

				// transform the character
				switch (toupper(ch)) {
					case 'A': ch = RTTY_A; break;
					case 'B': ch = RTTY_B; break;
					case 'C': ch = RTTY_C; break;
					case 'D': ch = RTTY_D; break;
					case 'E': ch = RTTY_E; break;
					case 'F': ch = RTTY_F; break;
					case 'G': ch = RTTY_G; break;
					case 'H': ch = RTTY_H; break;
					case 'I': ch = RTTY_I; break;
					case 'J': ch = RTTY_J; break;
					case 'K': ch = RTTY_K; break;
					case 'L': ch = RTTY_L; break;
					case 'M': ch = RTTY_M; break;
					case 'N': ch = RTTY_N; break;
					case 'O': ch = RTTY_O; break;
					case 'P': ch = RTTY_P; break;
					case 'Q': ch = RTTY_Q; break;
					case 'R': ch = RTTY_R; break;
					case 'S': ch = RTTY_S; break;
					case 'T': ch = RTTY_T; break;
					case 'U': ch = RTTY_U; break;
					case 'V': ch = RTTY_V; break;
					case 'W': ch = RTTY_W; break;
					case 'X': ch = RTTY_X; break;
					case 'Y': ch = RTTY_Y; break;
					case 'Z': ch = RTTY_Z; break;
					default: return 0;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the LS insertion, too

			} else { // some of the symbols
				switch (ch) {
					case '-': ch = RTTY_DASH; break;
					case '\'': ch = RTTY_QUOTE; break;
					case '$': ch = RTTY_DOLLAR; break;
					case '!': ch = RTTY_BANG; break;
					case '&': ch = RTTY_AMP; break;
					case '#': ch = RTTY_HASH; break;
					case '(': ch = RTTY_PARL; break;
					case ')': ch = RTTY_PARR; break;
					case '+': ch = RTTY_PLUS; break;
					case '/': ch = RTTY_SLASH; break;
					case ';': ch = RTTY_SEMI; break;
					case ':': ch = RTTY_COLON; break;
					case '?': ch = RTTY_QUES; break;
					case ',': ch = RTTY_COMMA; break;
					case '.': ch = RTTY_FSTOP; break;
					default: ch = 0; break;
				}
				if (!ch) {
#ifdef VERBOSE_DEBUG
					HostSerial.print("#debug: couldn't encode symbol;");
#endif
					// give up
					return 0;
				}

				int8_t bc = 0;
				if (txFresh || !txShift) {
					TxBuffer.Add((uint16_t)(RTTY_FS));
					txShift = true;
					txFresh = false;
					bc = 1;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the FS insertion, too
			}
			return false;
		case ASCII7:
			return TxBuffer.Add((uint16_t)((((~ch) & 0x7F) << 1) | 0x0001)) ? 1 : -1;
		case ASCII8:
			return TxBuffer.Add((uint16_t)(((~ch) << 1) | 0x0001)) ? 1 : -1;
		default:
			return false;
	}
}


////////////////////////////////////////////////////////////////
//
//  MAIN PROGRAM
//
////////////////////////////////////////////////////////////////

//
//  setup() - initialize hardware
//
void setup() {
	// start the serial port
	HostSerial.begin(SerialRate);

	// allocate strings
	cmdText.reserve(CommandStringLength);
	inputString.reserve(InputStringLength);
	inputString = strcpy_P(ioBuffer, PSTR("#VERSION;"));
	inputReady = true;

	// RTTY PTT pin
	#if RTTY_PTT_PIN > 0
	pinMode(RTTY_PTT_PIN, OUTPUT);
	digitalWrite(RTTY_PTT_PIN, LOW);
	pttState = false;
	#endif

	// RTTY KEY pin
	#if RTTY_KEY_PIN > 0
	pinMode(RTTY_KEY_PIN, OUTPUT);
	digitalWrite(RTTY_KEY_PIN, LOW);
	#endif

	// HB LED
	#if HEARTBEAT_LED > 0
	HeartbeatSetup();
	#endif

	#ifdef RTTY_TRIGRAPH
	trigraph.InnerBlankingEdge = TRI_LED_EDGE; // where do the outer LED's blank
	trigraph.OuterBlankingEdge = TRI_LED_EDGE; // where does the center LED blank
	trigraph.setup();
	trigraph.test();
	#endif

	// allocate the demodulator
	Demodulator::Config c;
	c.f0 = RTTY_DATA_FREQ_CENTER;
	c.fs = RTTY_SAMPLE_RATE;
	c.shift = RTTY_DATA_FREQ_SHIFT;
	c.R = RTTY_TONE_FILTER_R;
	c.int_len = RTTY_TONE_INTEGRATOR_LENGTH;
	c.trigger = RTTY_DISCRIMINATOR_HYSTERESIS_TRIGGER;
	demod = new Demodulator(c);

	// allocate audio memory
	AudioMemory(RTTY_AUDIO_BUFFER_FRAMES);

	// configure the input and output filters
	uint8_t order = RTTY_ORDER;
	if (order > 4) order = 4;
	for (uint8_t i = 0; i != order; ++i) {
		iirIn.setBandpass(i, RTTY_DATA_FREQ_CENTER, RTTY_DATA_FREQ_CENTER / RTTY_BW);
		//iirOut.setBandpass(0, RTTY_FREQ, RTTY_FREQ / RTTY_BW);
	}

	// read persisted settings
	SettingsBlock toRead;
	if (eeprom.read(reinterpret_cast<char*>(&toRead), sizeof(toRead)) == sizeof(toRead)) {
		Settings = toRead;
	} else {
		Settings.line_input = RTTY_INPUT_LEVEL;
	}

	// configure the shield for LINE INPUT, with no processing
	sgtl5000.enable();
	sgtl5000.inputSelect(AUDIO_INPUT_LINEIN);
	sgtl5000.lineInLevel(Settings.line_input);
	// configure the shield for LINE OUTPUT, with no processing
	sgtl5000.muteHeadphone();
	sgtl5000.unmuteLineout();
	sgtl5000.autoVolumeDisable();
	sgtl5000.surroundSoundDisable();
	sgtl5000.enhanceBassDisable();
	sgtl5000.eqSelect(FLAT_FREQUENCY);
	sgtl5000.adcHighPassFilterDisable();

	// compute bit lengths and masks
	ReconfigureTiming();

	// start sampling
	inQueue.begin();
}


//
//  loop() - main program loop
//
void loop() {
	static bool hash = false;

	// read the clock
	now = micros();

	// blink the HB LED
	HeartbeatLoop();

	// process PTT key-up delay
	if (!tx &&
		#if RTTY_PTT_PIN > 0
		pttState &&
		#endif
		Elapsed(now, rttyTxEdge) >= (Settings.pttLeadOut * 1000)) {
		#if RTTY_PTT_PIN > 0
		digitalWrite(RTTY_PTT_PIN, LOW);
		pttState = false;
		#endif
		// force pin LOW for RX
		delay(PTT_TO_UNKEY_DELAY);
		SetRttyState(LowTX()); // set to LOW when not transmitting, to extinquish LED

		// start sampling
		inQueue.begin();
	}
	
	// if we are in transmit mode, service the TX queue
	if (tx) {
		// re-read the clock
		now = micros();

		// if the current element has been sent...
		if (Elapsed(now, rttyTxEdge) >= bitPeriod) {
			bool bit;
			if (GetNextTxBit(bit)) {
				// generate the leading edge
				SetRttyState(bit);
			} else {
				tx = false;
				// go back to idle
				SetRttyState(HighTX());
			}
		}
	} // if (tx)...

	// if user input ready
	if (inputReady || hash) {
		// for each char received from the USB...
		for (size_t i = 0; hash || (i != inputString.length()); i += hash ? 0 : 1) {
			char ch = hash ? '#' : inputString.charAt(i);
			
			// if we are in command mode...
			if (cmd) {
				if (ch == '#') {
					// if a '#' sent mid-command, start over
					cmdText = "";
				} else if (ch == ';') {
					// terminate and process the command string
					cmd = false;
					bool ok = false;
					if (echo == 1) HostSerial.print(';');

					// parse the command and argument
					SplitCommand(cmdText, command, argument);
					if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOC")))) {
						//
						//  QUERY DOC
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += strcpy_P(ioBuffer, PSTR("=UOS,ECHO,MODE,REV,PTTIN,PTTOUT,SQL,SQTO,LINEIN,LEVEL,TRIGGER,HALFBIT,DEV,SFO,TONES,ANGLE,VERSION"));
						} else {
							//
							//  TODO: respond to individual doc requests
							//
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("CSET")))) {
						//
						//  QUERY/SET CHAR SET
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							switch (Settings.encoding) {
								case Baudot: cmdText += strcpy_P(ioBuffer, PSTR("BAUDOT")); break;
								case ASCII7: cmdText += strcpy_P(ioBuffer, PSTR("ASCII7")); break;
								case ASCII8: cmdText += strcpy_P(ioBuffer, PSTR("ASCII8")); break;
							}
						} else {
							if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BAUDOT")))) {
								Settings.encoding = Baudot;
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ASCII7")))) {
								Settings.encoding = ASCII7;
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ASCII8")))) {
								Settings.encoding = ASCII8;
								settingsDirty = true;
								ok = true;
							} else if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+':
										if (Settings.encoding < ASCII8) {
											Settings.encoding = (Encodings)(Settings.encoding + 1);
										}
										settingsDirty = true;
										ok = true;
										break;
									case '-':
										if (Settings.encoding > 0) {
											Settings.encoding = (Encodings)(Settings.encoding - 1);
										}
										settingsDirty = true;
										ok = true;
										break;
								}
								if (ok) {
									switch (Settings.encoding) {
										case Baudot: cmdText = command + strcpy_P(ioBuffer, PSTR("=BAUDOT")); break;
										case ASCII7: cmdText = command + strcpy_P(ioBuffer, PSTR("=ASCII7")); break;
										case ASCII8: cmdText = command + strcpy_P(ioBuffer, PSTR("=ASCII8")); break;
									}
								}
							}

							// recalculate masks
							ReconfigureTiming();
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("UOS")))) {
						//
						//  QUERY/SET RX UOS
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += uos ? "1" : "0";
						} else {
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+':
									case '-': {
										uos = uos ? false : true;
										cmdText = command + '=' + (int)(uos);
										settingsDirty = true;
										ok = true;
									} break;
									default: {
										int newUos = argument.toInt();
										uos = newUos ? true : false;
										settingsDirty = true;
										ok = true;
									} break;
								}
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVELMAX")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += 100;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVEL")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText = command + '=';

                        // read the peak and normalize to %; the peakIn will reset automatically
						cmdText += static_cast<uint16_t>(100.0 * peakIn.read());
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DEV")))) {
						// return the demod's current deviation value
						ok = true;
						cmdText += '=';
						cmdText += demod->deviation();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ANGLE")))) {
						// return the discriminator raw value
						ok = true;
						cmdText += '=';
						cmdText += demod->angle();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("SFO")))) {
						// return the demod's current deviation value
						ok = true;
						cmdText += '=';
						cmdText += demod->offset();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TONES")))) {
						// return the TPF LTA values
						ok = true;
						cmdText += '=';
						cmdText += demod->level_low();
						cmdText += ',';
						cmdText += demod->level_high();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DS")))) {
						// return the current input data sate
						ok = true;
						cmdText += '=';
						cmdText += rttyRxState;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("REV")))) {
						//
						//  QUERY/SET REVERSE POLARITY
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							switch (Settings.reverse) {
								case RevNone: cmdText += "NONE"; break;
								case RevRX: cmdText += "RX"; break;
								case RevTX: cmdText += "TX"; break;
								case RevBoth: cmdText += "BOTH"; break;
							}
						} else {
							if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("NONE")))) {
								Settings.reverse = RevNone;
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("RX")))) {
								Settings.reverse = RevRX;
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TX")))) {
								Settings.reverse = RevTX;
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BOTH")))) {
								Settings.reverse = RevBoth;
								settingsDirty = true;
								ok = true;
							} else if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (Settings.reverse < RevBoth) {
											Settings.reverse = (Reversals)(Settings.reverse + 1);
											settingsDirty = true;
											ok = true;
										}
									} break;
									case '-': {
										if (Settings.reverse > 0) {
											Settings.reverse = (Reversals)(Settings.reverse - 1);
											settingsDirty = true;
											ok = true;
										}
									} break;
								}
								if (ok) {
									switch (Settings.reverse) {
										case RevNone: cmdText = command + strcpy_P(ioBuffer, PSTR("=NONE")); break;
										case RevRX: cmdText = command + strcpy_P(ioBuffer, PSTR("=RX")); break;
										case RevTX: cmdText = command + strcpy_P(ioBuffer, PSTR("=TX")); break;
										case RevBoth: cmdText = command + strcpy_P(ioBuffer, PSTR("=BOTH")); break;
									}
								}
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ECHO")))) {
						//
						//  QUERY/SET ECHO
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(echo);
						} else {
							if (argument.length() == 1) {
								int newEcho = echo;
								switch (argument.charAt(0)) {
									case '+': {
										if (echo < 2)
											newEcho = echo + 1;
									} break;
									case '-': {
										if (echo > 0)
											newEcho = echo - 1;
									} break;
									default: {
										newEcho = argument.toInt();
									} break;
								}
								
								if (newEcho >= 0 && newEcho <= 2) {
									echo = newEcho;
									cmdText = command.toUpperCase() + '=' + echo;
									ok = true;
								}
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TUNE")))) {
						//
						//  SET TUNE
						//
						if ((argument.length() == 1) && !tx) {
							char doTune = argument.charAt(0);
							switch (doTune) {
								case '0':
									#if RTTY_PTT_PIN > 0
									// de-PTT the radio
									digitalWrite(RTTY_PTT_PIN, LOW);
									pttState = false;
									#endif

									// turn the receiver back on
									inQueue.begin();

									ok = true;
									break;
								case '1':
									// turn the receiver off
									inQueue.end();

									// set the frequency to the IDLE value
									digitalWrite(RTTY_KEY_PIN, HighTX());

									#if RTTY_PTT_PIN > 0
									// PTT the radio
									digitalWrite(RTTY_PTT_PIN, HIGH);
									pttState = false;
									#endif

									ok = true;
									break;
								default:
									// nop
									break;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LINEIN")))) {
						//
						//  QUERY/SET LINE IN VOLUME
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.line_input;
						} else {
							unsigned long newLevel = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newLevel = Settings.line_input;
										if (Settings.line_input < LINE_INPUT_MAX) {
											newLevel = Settings.line_input + 1;
										}
									} break;
									case '-': {
										newLevel = Settings.line_input;
										if (Settings.line_input > 0) {
											newLevel = Settings.line_input - 1;
										}
									} break;
								}
							}
							if (newLevel >= 0 && newLevel <= LINE_INPUT_MAX) {
								Settings.line_input = newLevel;
								sgtl5000.lineInLevel(newLevel);
								cmdText = command + '=' + Settings.line_input;
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("HALFBIT")))) {
						//
						//  QUERY/SET HALF BIT THRESHOLD (50 to 100%)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += static_cast<int>(round(100.0 * Settings.halfBitFactor));
						} else {
							unsigned long newBit = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newBit = static_cast<int>(100.0 * Settings.halfBitFactor);
										if (newBit < 100) {
											++newBit;
										}
									} break;
									case '-': {
										newBit = static_cast<int>(100.0 * Settings.halfBitFactor);
										if (newBit > 0) {
											--newBit;
										}
									} break;
								}
							}
							if (newBit >= 10 && newBit <= 100) {
								Settings.halfBitFactor = static_cast<double>(newBit) / 100.0;
								cmdText = command + '=' + static_cast<int>(round(100.0 * Settings.halfBitFactor));
								halfBit = (Settings.halfBitFactor * bitPeriod);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TRIGGER")))) {
						//
						//  QUERY/SET HYSTERESIS TRIGGER (0 to 100%)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += static_cast<int>(round(100.0 * demod->trig_level()));
						} else {
							unsigned long newLevel = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newLevel = static_cast<int>(100.0 * demod->trig_level());
										if (newLevel < 100) {
											++newLevel;
										}
									} break;
									case '-': {
										newLevel = static_cast<int>(100.0 * demod->trig_level());
										if (newLevel > 0) {
											--newLevel;
										}
									} break;
								}
							}
							if (newLevel >= 0 && newLevel <= 100) {
								demod->trig_level(static_cast<double>(newLevel) / 100.0);
								//
								// TODO: store in Settings block, too
								//
								cmdText = command + '=' + static_cast<int>(round(100.0 * demod->trig_level()));
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("SQTO")))) {
						//
						//  QUERY/SET SQUELCH TIMEOUT
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.squelch_timeout;
						} else {
							unsigned long newSql = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newSql = Settings.squelch_timeout;
										if (Settings.squelch_timeout < SQUELCH_TIMEOUT_MAX) {
											newSql = Settings.squelch_timeout + 1;
										}
									} break;
									case '-': {
										newSql = Settings.squelch_timeout;
										if (Settings.squelch_timeout > 0) {
											newSql = Settings.squelch_timeout - 1;
										}
									} break;
								}
							}
							if (newSql >= 0 && newSql <= SQUELCH_TIMEOUT_MAX) {
								Settings.squelch_timeout = newSql;
								cmdText = command + '=' + Settings.squelch_timeout;
								ok = true;
								settingsDirty = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("SQL")))) {
						//
						//  QUERY/SET SQUELCH
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.squelch;
						} else {
							unsigned long newSql = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newSql = Settings.squelch;
										if (Settings.squelch < KK5JY::DSP::norm_limits<int8_t>::maximum) {
											newSql = Settings.squelch + 1;
										}
									} break;
									case '-': {
										newSql = Settings.squelch;
										if (Settings.squelch > 0) {
											newSql = Settings.squelch - 1;
										}
									} break;
								}
							}
							if (newSql >= 0 && newSql <= static_cast<uint8_t>(KK5JY::DSP::norm_limits<int8_t>::maximum)) {
								Settings.squelch = newSql;
								cmdText = command + '=' + Settings.squelch;
								ok = true;
								settingsDirty = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTIN")))) {
						//
						//  QUERY/SET PTT LEAD-IN
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.pttLeadIn;
						} else {
							unsigned long newPTT = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newPTT = Settings.pttLeadIn;
										if (Settings.pttLeadIn < MaxPttLeadIn) {
											newPTT = Settings.pttLeadIn + 1;
										}
									} break;
									case '-': {
										newPTT = Settings.pttLeadIn;
										if (Settings.pttLeadIn > 0) {
											newPTT = Settings.pttLeadIn - 1;
										}
									} break;
								}
							}
							if (newPTT >= 0 && newPTT <= MaxPttLeadIn) {
								Settings.pttLeadIn = newPTT;
								cmdText = command + '=' + Settings.pttLeadIn;
								ok = true;
								settingsDirty = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTOUT")))) {
						//
						//  QUERY/SET PTT LEAD-OUT
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.pttLeadOut;
						} else {
							unsigned long newPTT = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newPTT = Settings.pttLeadOut;
										if (Settings.pttLeadOut < MaxPttLeadOut) {
											newPTT = Settings.pttLeadOut + 1;
										}
									} break;
									case '-': {
										newPTT = Settings.pttLeadOut;
										if (Settings.pttLeadOut > 0) {
											newPTT = Settings.pttLeadOut - 1;
										}
									} break;
								}
							}
							if (newPTT >= 0 && newPTT <= MaxPttLeadOut) {
								Settings.pttLeadOut = newPTT;
								cmdText = command + '=' + Settings.pttLeadOut;
								ok = true;
								settingsDirty = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("MODE")))) {
						//
						//  QUERY MODEM MODE
						//
						ok = true;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR("RTTY"));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("VERSION")))) {
						//
						//  QUERY VERSION NUMBER
						//
						ok = true;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR(VERSION_STRING));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("HASH")))) {
						//
						//  SEND HASH
						//
						hash = true;
						ok = true;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("EC")))) {
						//
						//  CLEAR EEPROM
						//
						if (argument.length() == 1 && argument.charAt(0) == '1') {
							eeprom.clear();
							settingsDirty = false;
							ok = true;
						}
					}
					
					//
					//  TODO: process other command contents
					//

					// respond with #OK or #ERR
					if (echo == 1) {
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					}

					if (command.length() == 0) {
						HostSerial.print(strcpy_P(ioBuffer, PSTR("#OK")));
					} else {
						HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK:")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
						HostSerial.print(cmdText);
					}
					HostSerial.print(';');
					if (echo == 1) {
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					}
					cmdText = "";
				} else {
					if (isAlpha(ch) || isDigit(ch) || isPunct(ch) || (ch == ' ')) {
						// add the char to the command string
						if (cmdText.length() < CommandStringLength)
							cmdText += ch;
						
						// echo the char, if enabled
						if (echo == 1) {
							HostSerial.print(ch);
						}
					}
				}
				continue;
			}
			
			// start command mode??
			if (!hash) {
				if ((!cmd) && (ch == '#')) {
					cmd = true;
					if (echo == 1) HostSerial.print('#');
					continue;
				}
			}

			// hard-stop?
			if (ch == ESC || ch == CtlC) {
				#if RTTY_PTT_PIN > 0
				digitalWrite(RTTY_PTT_PIN, LOW);
				pttState = false;
				#endif
				delay(PTT_TO_UNKEY_DELAY);
				SetRttyState(LowTX()); // set LOW when not transmitting to extinquish LED
				TxBuffer.Clear();
				tx = false;

				// start sampling
				inQueue.begin();

				continue;
			}

			// echo the char, if enabled
			if (hash) {
				hash = false;
			} else if (echo) {
				HostSerial.print(ch);
			}

			//
			//  TX - drive the output pin and PTT
			//
			int8_t eok = Encode(ch);
			if (eok < 0) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#OVF;")));
			}
			if (eok && !tx) {
				bool bit;
				if (GetNextTxBit(bit)) {
					// stop sampling
					inQueue.end();

					#ifdef RTTY_TRIGRAPH
					// mute the RX trigraph
					trigraph.clear();
					#endif

					// generate idle
					SetRttyState(HighTX());

					// PTT lead-in
					#if RTTY_PTT_PIN > 0
					pttState = true;
					digitalWrite(RTTY_PTT_PIN, HIGH);
					#endif

					#if USE_PTT_DELAY
					delay(Settings.pttLeadIn);
					#else
					unsigned long pttStart = millis();
					while (Elapsed(millis(), pttStart) <= Settings.pttLeadIn) {
						// keep the RX queue from overrunning
						char ch = ServiceSerialPortInputQueue();

						// if hard-stop was pressed, just exit now so those handlers can run
						if (ch == ESC || ch == CtlC) {
							break;
						}
					}
					#endif

					// start sending
					tx = true;
					txFresh = true;

					// generate the leading edge
					SetRttyState(bit);
				}
			} // Encode(...) && !tx
		}

		// clear the string:
		inputString = "";
		inputReady = false;
	}

	//
	//  RX - process incoming bits
	//
	if (!tx) { // if not transmitting...
		// service incoming AFSK data
		ProcessRxQueue();
	}

	// update the EEPROM if settings have changed
	uint32_t ms = millis();
	if (settingsDirty) {
		if (Elapsed(ms, lastSettingsWrite) > SettingsInterval) {
			eeprom.write(reinterpret_cast<char*>(&Settings), sizeof(Settings));
			settingsDirty = false;
			lastSettingsWrite = ms;
		}
	}

	// close the squelch if it has timed out
	if (!squelch) {
		if (Elapsed(ms, squelchOpened) > Settings.squelch_timeout) {
			squelch = true;
		}
	}

	// service the host port
	ServiceSerialPortInputQueue();
}

// EOF
